@@ -577,7 +577,7 @@ add_struct_block_before(fz_context *ctx, fz_stext_block *before, fz_stext_page *
 	if (before)
 	{
 		int idx2 = idx+1;
-		for (block = before->next; block != NULL; block = block->next)
+		for (block = before; block != NULL; block = block->next)
 		{
 			if (block->type != FZ_STEXT_BLOCK_STRUCT)
 				continue;
@@ -2303,6 +2303,7 @@ move_contained_content(fz_context *ctx, fz_stext_page *page, fz_stext_struct *de
 	fz_stext_block **sfirst = src ? &src->first_block : &page->first_block;
 	fz_stext_block **slast = src ? &src->last_block : &page->last_block;
 	fz_stext_block *block, *next_block;
+	int max;
 
 	for (block = *sfirst; block != NULL; block = next_block)
 	{
@@ -2322,6 +2323,7 @@ move_contained_content(fz_context *ctx, fz_stext_page *page, fz_stext_struct *de
 					move_contained_content(ctx, page, dest, block->u.s.down, r);
 					continue;
 				}
+				block->u.s.down->parent = dest;
 			}
 			unlink_block(block, sfirst, slast);
 			insert_block_before(block, before, page, dest);
@@ -2466,6 +2468,19 @@ move_contained_content(fz_context *ctx, fz_stext_page *page, fz_stext_struct *de
 		}
 	}
 
+	/* Now run through where we put blocks ensuring the indexes make sense. */
+	max = 0;
+	before = dest ? dest->first_block : page->first_block;
+	for (block = before; block != NULL; block = block->next)
+	{
+		if (block->type != FZ_STEXT_BLOCK_STRUCT)
+			continue;
+
+		if (block->u.s.index < max)
+			block->u.s.index = max;
+		max = block->u.s.index+1;
+	}
+
 	return r;
 }
 
@@ -2552,12 +2567,19 @@ tidy_td_divs(fz_context *ctx, fz_stext_struct *parent)
 	while (1)
 	{
 		fz_stext_block *block = parent->first_block;
+		fz_stext_block *b;
 
 		if (block == NULL || block->next != NULL || block->type != FZ_STEXT_BLOCK_STRUCT || block->u.s.down == NULL || block->u.s.down->standard != FZ_STRUCTURE_DIV)
 			return;
 
 		parent->first_block = block->u.s.down->first_block;
 		parent->last_block = block->u.s.down->last_block;
+
+		for (b = parent->first_block; b != NULL; b = b->next)
+		{
+			if (b->type == FZ_STEXT_BLOCK_STRUCT && b->u.s.down)
+				b->u.s.down->parent = parent;
+		}
 	}
 }
 
