@@ -1,7 +1,11 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#ifdef _MSC_VER
+#include <io.h>
+#else
 #include <unistd.h>
+#endif
 #include <errno.h>
 
 #include "mujs.h"
@@ -199,6 +203,11 @@ static void jsB_quit(js_State *J)
 	exit(js_tonumber(J, 1));
 }
 
+static void jsB_repr(js_State *J)
+{
+	js_repr(J, 1);
+}
+
 static const char *require_js =
 	"function require(name) {\n"
 	"var cache = require.cache;\n"
@@ -211,16 +220,23 @@ static const char *require_js =
 	"require.cache = Object.create(null);\n"
 ;
 
+
 static const char *stacktrace_js =
 	"Error.prototype.toString = function() {\n"
-	"if (this.stackTrace) return this.name + ': ' + this.message + this.stackTrace;\n"
-	"return this.name + ': ' + this.message;\n"
+	"var s = this.name;\n"
+	"if ('message' in this) s += ': ' + this.message;\n"
+	"if ('stack' in this) s += this.stack;\n"
+	"return s;\n"
 	"};\n"
 ;
 
+static const char *console_js =
+	"var console = { log: print, debug: print, warn: print, error: print };"
+;
+
 static int eval_print(js_State *J, const char *source)
 {
-	if (js_ploadstring(J, "[string]", source)) {
+	if (js_ploadstring(J, "[stdin]", source)) {
 		fprintf(stderr, "%s\n", js_trystring(J, -1, "Error"));
 		js_pop(J, 1);
 		return 1;
@@ -231,8 +247,9 @@ static int eval_print(js_State *J, const char *source)
 		js_pop(J, 1);
 		return 1;
 	}
-	if (js_isdefined(J, -1))
-		printf("%s\n", js_trystring(J, -1, "can't convert to string"));
+	if (js_isdefined(J, -1)) {
+		printf("%s\n", js_tryrepr(J, -1, "can't convert to string"));
+	}
 	js_pop(J, 1);
 	return 0;
 }
@@ -283,17 +300,25 @@ main(int argc, char **argv)
 	int status = 0;
 	int strict = 0;
 	int interactive = 0;
+	int runlimit = 0;
+	int memlimit = 0;
 	int i, c;
 
-	while ((c = xgetopt(argc, argv, "is")) != -1) {
+	while ((c = xgetopt(argc, argv, "isR:M:")) != -1) {
 		switch (c) {
 		default: usage(); break;
 		case 'i': interactive = 1; break;
 		case 's': strict = 1; break;
+		case 'R': runlimit = atoi(xoptarg); break;
+		case 'M': memlimit = atoi(xoptarg); break;
 		}
 	}
 
 	J = js_newstate(NULL, NULL, strict ? JS_STRICT : 0);
+	if (!J) {
+		fprintf(stderr, "Could not initialize MuJS.\n");
+		exit(1);
+	}
 
 	js_newcfunction(J, jsB_gc, "gc", 0);
 	js_setglobal(J, "gc");
@@ -316,11 +341,15 @@ main(int argc, char **argv)
 	js_newcfunction(J, jsB_readline, "readline", 0);
 	js_setglobal(J, "readline");
 
+	js_newcfunction(J, jsB_repr, "repr", 0);
+	js_setglobal(J, "repr");
+
 	js_newcfunction(J, jsB_quit, "quit", 1);
 	js_setglobal(J, "quit");
 
 	js_dostring(J, require_js);
 	js_dostring(J, stacktrace_js);
+	js_dostring(J, console_js);
 
 	if (xoptind == argc) {
 		interactive = 1;
@@ -335,16 +364,20 @@ main(int argc, char **argv)
 		}
 		js_setglobal(J, "scriptArgs");
 
+		js_setlimit(J, runlimit, memlimit);
 		if (js_dofile(J, argv[c]))
 			status = 1;
 	}
 
 	if (interactive) {
+		printf("Welcome to MuJS %d.%d.%d.\n",
+			JS_VERSION_MAJOR, JS_VERSION_MINOR, JS_VERSION_PATCH);
 		if (isatty(0)) {
 			using_history();
 			rl_bind_key('\t', rl_insert);
 			input = readline(PS1);
 			while (input) {
+				js_setlimit(J, runlimit, memlimit);
 				eval_print(J, input);
 				if (*input)
 					add_history(input);
@@ -354,6 +387,7 @@ main(int argc, char **argv)
 			putchar('\n');
 		} else {
 			input = read_stdin();
+			js_setlimit(J, runlimit, memlimit);
 			if (!input || !js_dostring(J, input))
 				status = 1;
 			free(input);
